#define	je_pool_create JEMALLOC_N(pool_create)
#define	je_pool_delete JEMALLOC_N(pool_delete)
#define	je_pool_malloc JEMALLOC_N(pool_malloc)
#define	je_pool_calloc JEMALLOC_N(pool_calloc)
#define	je_pool_ralloc JEMALLOC_N(pool_ralloc)
#define	je_pool_aligned_alloc JEMALLOC_N(pool_aligned_alloc)
#define	je_pool_free JEMALLOC_N(pool_free)
#define	je_pool_malloc_usable_size JEMALLOC_N(pool_malloc_usable_size)
#define	je_pool_malloc_stats_print JEMALLOC_N(pool_malloc_stats_print)
#define	je_pool_extend JEMALLOC_N(pool_extend)
#define	je_pool_set_alloc_funcs JEMALLOC_N(pool_set_alloc_funcs)
#define	je_pool_check JEMALLOC_N(pool_check)
#define	je_malloc_conf JEMALLOC_N(malloc_conf)
#define	je_malloc_message JEMALLOC_N(malloc_message)
#define	je_malloc JEMALLOC_N(malloc)
#define	je_calloc JEMALLOC_N(calloc)
#define	je_posix_memalign JEMALLOC_N(posix_memalign)
#define	je_aligned_alloc JEMALLOC_N(aligned_alloc)
#define	je_realloc JEMALLOC_N(realloc)
#define	je_free JEMALLOC_N(free)
#define	je_mallocx JEMALLOC_N(mallocx)
#define	je_rallocx JEMALLOC_N(rallocx)
#define	je_xallocx JEMALLOC_N(xallocx)
#define	je_sallocx JEMALLOC_N(sallocx)
#define	je_dallocx JEMALLOC_N(dallocx)
#define	je_nallocx JEMALLOC_N(nallocx)
#define	je_mallctl JEMALLOC_N(mallctl)
#define	je_mallctlnametomib JEMALLOC_N(mallctlnametomib)
#define	je_mallctlbymib JEMALLOC_N(mallctlbymib)
#define	je_navsnprintf JEMALLOC_N(navsnprintf)
#define	je_malloc_stats_print JEMALLOC_N(malloc_stats_print)
#define	je_malloc_usable_size JEMALLOC_N(malloc_usable_size)
