#
# Copyright 2017, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# ps_analyze -- script to analyze ps1 files
#

Write-Output "Starting PSScript analyzing ..."

$scriptdir = Split-Path -Parent $PSCommandPath
$rootdir = $scriptdir + "\.."
$detected = 0

$include = @("*.ps1" )
Get-ChildItem -Path $rootdir -Recurse -Include $include  | `
    Where-Object { $_.FullName -notlike "*test*" } | `
    ForEach-Object {
        $analyze_result = Invoke-ScriptAnalyzer -Path $_.FullName
        if ($analyze_result) {
            $detected = $detected + $analyze_result.Count
            Write-Output $_.FullName
            Write-Output $analyze_result
        }
    }

if ($detected) {
    Write-Output "PSScriptAnalyzer FAILED. Issues detected: $detected"
    Exit 1
} else {
    Write-Output "PSScriptAnalyzer PASSED. No issue detected."
    Exit 0
}
