#
# Copyright 2016-2017, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# CSTYLE.ps1 -- script to check coding style
#
# XXX - integrate with VS projects and execute for each build
#

$scriptdir = Split-Path -Parent $PSCommandPath
$rootdir = $scriptdir + "\.."
$cstyle = $rootdir + "\utils\cstyle"
$checkdir = $rootdir

# XXX - *.cpp/*.hpp files not supported yet
$include = @( "*.c", "*.h" )

If ( Get-Command -Name perl -ErrorAction SilentlyContinue ) {
	Get-ChildItem -Path $checkdir -Recurse -Include $include | `
    Where-Object { $_.FullName -notlike "*jemalloc*" } | `
    ForEach-Object {
        $IGNORE = $_.DirectoryName + "\.cstyleignore"
        if(Test-Path $IGNORE) {
            if((Select-String $_.Name $IGNORE)) {
                return
            }
        }
        $_
    } | ForEach-Object {
		Write-Output $_.FullName
		& perl $cstyle $_.FullName
		if ($LASTEXITCODE -ne 0) {
            Exit $LASTEXITCODE
        }
    }
} else {
	Write-Output "Cannot execute cstyle - perl is missing"
}
