#
# Copyright 2015-2018, Intel Corporation
# Copyright (c) 2016, Microsoft Corporation. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# src/test/blk_pool/TEST30 -- unit test for pmemblk_create
#

# standard unit test setup
. ..\unittest\unittest.ps1

require_test_type medium

setup

#
# TEST30 existing file, file size >= min required size, bsize == 300
#        (pool set with replica section)
#
# XXX fix one we figure out windows poolsets and $PMEMSPOIL tool(?)
echo "XXX" > out30w.log

#expect_normal_exit $Env:EXE_DIR\blk_pool$Env:EXESUFFIX `
#    c $DIR\testfile1 300 20 0640

#cp $DIR\testfile1 $DIR\testfile2

#$PMEMSPOIL $DIR\testfile1 pool_hdr.uuid=1111111111111111 \
#	pool_hdr.next_part_uuid=1111111111111111 \
#	pool_hdr.prev_part_uuid=1111111111111111 \
#	pool_hdr.next_repl_uuid=2222222222222222 \
#	pool_hdr.prev_repl_uuid=2222222222222222 \
#	"pool_hdr.checksum_gen()"

#$PMEMSPOIL $DIR\testfile2 pool_hdr.uuid=2222222222222222 \
#	pool_hdr.next_part_uuid=2222222222222222 \
#	pool_hdr.prev_part_uuid=2222222222222222 \
#	pool_hdr.next_repl_uuid=1111111111111111 \
#	pool_hdr.prev_repl_uuid=1111111111111111 \
#	"pool_hdr.checksum_gen()"

#create_poolset $DIR\testset 20M:$DIR\testfile1 R 20M:$DIR\testfile2

#expect_normal_exit $Env:EXE_DIR\blk_pool$Env:EXESUFFIX `
#    c $DIR\testset 300

check

pass
